/***************************************************************************
 *   Copyright (C) 2005 by Roberto Cappuccio                               *
 *   robertoc@dnet.it                                                      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <kdebug.h>
#include <kglobal.h>
#include <qapplication.h>
#include <qtimer.h>
#include <qdir.h>
#include <kio/global.h>
#include <kio/jobclasses.h>
#include <kio/metainfojob.h>
#include <kio/previewjob.h>
#include <qbuffer.h>
#include <qregexp.h>

#include "kat.h"
#include "katcatalog.h"
#include "katinfoextractor.h"
#include "katfulltextjob.h"
#include "katpreviewjob.h"
#include "dlgmainwindow.h"

#define LIMIT 10

using namespace KIO;

KatInfoExtractor::KatInfoExtractor( KatCatalog* catalog, KatDatabaseManager *_dm )
    : KIO::Job( false ),
                m_jobs( 0L ), m_catalog( catalog ), m_bAsync( true ), m_totalSize( 0L ), m_totalFiles( 0L ),
                m_totalSubdirs( 0L ), m_fileTableRow( 0L ), m_totalExtractions( 0L ), m_totalPreviews( 0L ),
                m_totalMetaData( 0L ), m_totalFullTexts( 0L ), dm( _dm )
{
    startNextJob();
}

void KatInfoExtractor::startNextJob() {
    // read next LIMIT filenames from the database
    QMap<QString, long> files = dm->readFiles( m_fileTableRow, LIMIT );

    if ( files.empty() ) {
        // no more items in the database
        emit completed();
        if ( !m_bAsync ) {
            qApp->exit_loop();
        }
        // call the main job's slotResult
        KIO::Job::slotResult( this );
        return;
    }

    // create the 3 subjobs for the LIMIT files
    KFileItemList kfiles_preview;
    KFileItemList kfiles_metainfo;
    KFileItemList kfiles_fulltext;

    QMap<QString,long>::Iterator it;
    for ( it = files.begin(); it != files.end(); ++it ) {
        QString fullName = it.key();
        long fileId = it.data();

        kdDebug() << fullName << " : " << endl;

        KFileItem* kfi_preview = new KFileItem( KFileItem::Unknown, // _mode
                                                KFileItem::Unknown, // _permissions
                                                KURL( fullName ), // url
                                                true // determineMimeTypeOnDemand
                                               );
        KFileItem* kfi_metainfo = new KFileItem( *kfi_preview );
        KFileItem* kfi_fulltext = new KFileItem( *kfi_preview );

        if ( kfi_preview->mimetype() != "inode/directory" ) {
            extraInfo* ei_preview = new extraInfo();
            extraInfo* ei_metainfo = new extraInfo();
            extraInfo* ei_fulltext = new extraInfo();

            ei_preview->fileId = fileId;
            ei_metainfo->fileId = fileId;
            ei_fulltext->fileId = fileId;

            kfi_preview->setExtraData( "extra", ei_preview );
            kfi_metainfo->setExtraData( "extra", ei_metainfo );
            kfi_fulltext->setExtraData( "extra", ei_fulltext );

            kfiles_preview.append( kfi_preview );
            kfiles_metainfo.append( kfi_metainfo );
            kfiles_fulltext.append( kfi_fulltext );
        }
    }

    if ( !kfiles_preview.isEmpty() ) {

        KatPreviewJob* pj = new KatPreviewJob( kfiles_preview, // items
                                               m_catalog->getThumbnailSize(), // width
                                               m_catalog->getThumbnailSize(), // height
                                               0, // iconSize
                                               0, // iconAlpha
                                               false, // scale
                                               NULL, // enabledPlugins TODO: implement
                                               true // deleteItems
                                              );
        this->addSubjob( pj );

        connect( pj, SIGNAL( gotPreview( const KFileItem*, const QPixmap& ) ),
                    this, SLOT( gotThumbnail( const KFileItem*, const QPixmap& ) ) );
        connect( pj, SIGNAL( failed( const KFileItem* ) ),
                    this, SLOT( noThumbnail( const KFileItem* ) ) );
        connect( pj, SIGNAL( result( KIO::Job* ) ),
                    this, SLOT( resultThumbnail( KIO::Job* ) ) );
        m_jobs++;
    }

    if ( !kfiles_metainfo.isEmpty() ) {

        KIO::MetaInfoJob* mij = new KIO::MetaInfoJob( kfiles_metainfo, // items
                                                      true // deleteItems
                                                     );
        this->addSubjob( mij );

        connect( mij, SIGNAL( gotMetaInfo( const KFileItem* ) ),
                    this, SLOT( gotMetaInfo( const KFileItem* ) ) );
        connect( mij, SIGNAL( failed( const KFileItem* ) ),
                    this, SLOT( noMetaInfo( const KFileItem* ) ) );
        connect( mij, SIGNAL( result( KIO::Job* ) ),
                    this, SLOT( resultMetaInfo( KIO::Job* ) ) );
        m_jobs++;
    }

    if ( !kfiles_fulltext.isEmpty() ) {

        KatFullTextJob* ftj = new KatFullTextJob( kfiles_fulltext, // items
                                                  NULL, // enabledPlugins TODO: implement
                                                  true // deleteItems
                                                 );
        this->addSubjob( ftj );

        connect( ftj, SIGNAL( gotFullText( KIO::Job*, const KFileItem*, const QString& ) ),
                    this, SLOT( gotFullText( KIO::Job*, const KFileItem*, const QString& ) ) );
        connect( ftj, SIGNAL( failed( const KFileItem* ) ),
                    this, SLOT( noFullText( const KFileItem* ) ) );
        connect( ftj, SIGNAL( result( KIO::Job* ) ),
                    this, SLOT( resultFullText( KIO::Job* ) ) );
        m_jobs++;
    }

    m_fileTableRow += LIMIT;

    if ( m_jobs == 0L ) {
        // the database-items in the LIMIT were all directories,
        // so no job has been created.
        // directly start the next job
        startNextJob();
    }
}

KatInfoExtractor* KatInfoExtractor::extractInfoJob( KatCatalog* catalog, KatDatabaseManager *dm ) {
    return new KatInfoExtractor( catalog, dm );
}

void KatInfoExtractor::slotResult( KIO::Job* job ) {
    // slotResult is called whenever a subjob finishes (job is the subjob)
    subjobs.remove( job );
}

void KatInfoExtractor::gotThumbnail( const KFileItem* kfileitem, const QPixmap& qpixmap ) {
    extraInfo* extra = (extraInfo*)kfileitem->extraData( "extra" );
    long fileid = extra->fileId;
    delete extra;
    extra = 0L;

    QByteArray ba;
    QBuffer buffer( ba );
    buffer.open( IO_WriteOnly );
    qpixmap.save( &buffer, "PNG" );
    buffer.close();

    dm->saveThumbnailRecord( fileid, ba, ba.size() );

    m_totalPreviews++;
    m_catalog->setThumbnails( m_totalPreviews );

    m_totalExtractions++;
    m_catalog->setExtractions( m_totalExtractions );
}

void KatInfoExtractor::noThumbnail( const KFileItem* kfileitem ) {
    m_totalExtractions++;
    m_catalog->setExtractions( m_totalExtractions );
}

void KatInfoExtractor::gotMetaInfo( const KFileItem* kfileitem ) {
    extraInfo* extra = (extraInfo*)kfileitem->extraData( "extra" );
    long fileid = extra->fileId;
    delete extra;
    extra = 0L;

    QString metaData = "";

    if ( kfileitem->metaInfo().isValid() && !kfileitem->metaInfo().isEmpty() ) {

        QStringList sk = kfileitem->metaInfo().supportedKeys();

        for ( QStringList::Iterator it = sk.begin(); it != sk.end(); ++it ) {
            QString key = *it;

            if ( kfileitem->metaInfo().contains( key ) ) {
                // PRESENT

                KFileMetaInfoItem mii = kfileitem->metaInfo().item( key );

                if ( mii.isValid() ) {
                    QString mKey( mii.key() );

                    QVariant::Type type( mii.type() );
                    QString sType;

                    switch ( type ) {
                        case QVariant::BitArray:
                            sType = "bitarray";
                            break;
                        case QVariant::ByteArray:
                            sType = "bytearray";
                            break;
                        case QVariant::Bitmap:
                            sType = "bitmap";
                            break;
                        case QVariant::Bool:
                            sType = "bool";
                            break;
                        case QVariant::Brush:
                            sType = "brush";
                            break;
                        case QVariant::Color:
                            sType = "color";
                            break;
                        case QVariant::ColorGroup:
                            sType = "colorgroup";
                            break;
                        case QVariant::Cursor:
                            sType = "cursor";
                            break;
                        case QVariant::Date:
                            sType = "date";
                            break;
                        case QVariant::DateTime:
                            sType = "datetime";
                            break;
                        case QVariant::Double:
                            sType = "double";
                            break;
                        case QVariant::Font:
                            sType = "font";
                            break;
                        case QVariant::IconSet:
                            sType = "iconset";
                            break;
                        case QVariant::Image:
                            sType = "image";
                            break;
                        case QVariant::Int:
                            sType = "int";
                            break;
                        case QVariant::KeySequence:
                            sType = "keysequence";
                            break;
                        case QVariant::List:
                            sType = "list";
                            break;
                        case QVariant::LongLong:
                            sType = "longlong";
                            break;
                        case QVariant::ULongLong:
                            sType = "unsignedlonglong";
                            break;
                        case QVariant::Map:
                            sType = "map";
                            break;
                        case QVariant::Palette:
                            sType = "palette";
                            break;
                        case QVariant::Pen:
                            sType = "pen";
                            break;
                        case QVariant::Pixmap:
                            sType = "pixmap";
                            break;
                        case QVariant::Point:
                            sType = "point";
                            break;
                        case QVariant::PointArray:
                            sType = "pointarray";
                            break;
                        case QVariant::Rect:
                            sType = "rect";
                            break;
                        case QVariant::Region:
                            sType = "region";
                            break;
                        case QVariant::Size:
                            sType = "size";
                            break;
                        case QVariant::SizePolicy:
                            sType = "sizepolicy";
                            break;
                        case QVariant::String:
                            sType = "string";
                            break;
                        case QVariant::CString:
                            sType = "cstring";
                            break;
                        case QVariant::StringList:
                            sType = "stringlist";
                            break;
                        case QVariant::Time:
                            sType = "time";
                            break;
                        case QVariant::UInt:
                            sType = "unsignedint";
                            break;
                        default:
                            sType = "unknown";
                    }

                    if ( QString( mii.string() ) == "" ) {
                        // INVALID VALUE
                    } else {

                        QString mValue( mii.string() );
                        mValue = mValue.simplifyWhiteSpace();

                        if ( mValue != "" ) {
                            metaData = metaData + mKey + "|" + sType + "|" + mValue + "|";
                        }
                    }
                } else {
                    // INVALID ITEM
                }
            } else {
                // ABSENT
            }
        }

        if ( metaData != "" ) {
            dm->saveMetaDataRecord( fileid, metaData );

            m_totalMetaData++;
            m_catalog->setMetaData( m_totalMetaData );
        }

    } else {
        // INVALID OR EMPTY METAINFO
    }

    m_totalExtractions++;
    m_catalog->setExtractions( m_totalExtractions );
}

void KatInfoExtractor::noMetaInfo( const KFileItem* kfileitem ) {
    kdDebug() << "noMetaInfo" << endl;
    m_totalExtractions++;
    m_catalog->setExtractions( m_totalExtractions );
}

void KatInfoExtractor::gotFullText( KIO::Job* job, const KFileItem* kfileitem, const QString& fullText ) {
    extraInfo* extra = (extraInfo*)kfileitem->extraData( "extra" );
    long fileId = extra->fileId;
    delete extra;
    extra = 0L;

    QString text( fullText );
    if ( text == "<fulltext></fulltext>" ) {
        m_totalFullTexts++;
        m_catalog->setFullTexts( m_totalFullTexts );
        m_totalExtractions++;
        m_catalog->setExtractions( m_totalExtractions );
        return;
    }

    text = text.replace( "<fulltext>", QString::null ); // strip the <fulltext></fulltext> tags
    text = text.replace( "</fulltext>", QString::null );
    text = text.replace( QRegExp( "<page[^>]*>" ), " " ); // strip all <page x> tags
    text = text.replace( QRegExp( "</page>" ), " " ); // strip all </page> tags
    text = text.simplifyWhiteSpace(); // clean up white spaces

    if ( text.isEmpty() ) {
        m_totalFullTexts++;
        m_catalog->setFullTexts( m_totalFullTexts );
        m_totalExtractions++;
        m_catalog->setExtractions( m_totalExtractions );
        return;
    }

    // save fulltext
    QByteArray fullTextData = QCString( text );
    fullTextData = qCompress( fullTextData );
    long fullTextDataLength = fullTextData.size();
    dm->saveFullTextRecord( fileId, fullTextData, fullTextDataLength );

    // clean up the text a bit more
    text = text.lower();
    text = text.replace( QRegExp( "[\\W]" ), " " );
    text = text.replace( QRegExp( "[0-9]" ), " " );
    text = text.simplifyWhiteSpace();

    if ( text.isEmpty() ) {
        m_totalFullTexts++;
        m_catalog->setFullTexts( m_totalFullTexts );
        m_totalExtractions++;
        m_catalog->setExtractions( m_totalExtractions );
        return;
    }

    // create a list of words out of fulltext
    QStringList words = QStringList::split( " ", text, false );
    if ( !words.empty() ) {
        words.sort();

        long wordId;
        long o;

        // make all operations in one single transaction
        dm->openTransaction();

        // cycle the words, count occurrences and check if they are
        // already present in database
        long wordCount = 0L;
        long wordsTotal = words.size();
        QStringList::Iterator it = words.begin();
        while ( it != words.end() ) {

            QString currentWord = *it;
            it++;

            o = 1;
            while ( *it == currentWord ) {
                o++;
                it++;
            }

            // check if the word is already present in the database
            wordId = dm->getWordId( currentWord );
            if ( wordId == -1 ) {
                // if the word is not present, save it with a new wordid
                wordId = dm->saveWordRecord( currentWord );
                dm->saveWordfileRecord( wordId, fileId, o );

                m_catalog->setWords( m_catalog->getWords() + 1 );
            } else {
                // if it is present, save only the wordfile record
                dm->saveWordfileRecord( wordId, fileId, o );
            }
        }
        dm->commitTransaction();
    }

    if ( text.length() < 20 ) { // If it is shorter, we don't try to identify language
        m_totalFullTexts++;
        m_catalog->setFullTexts( m_totalFullTexts );
        m_totalExtractions++;
        m_catalog->setExtractions( m_totalExtractions );
        return;
    }

    // identify the language of the document
    QString language = KatLanguageManager::identifyLanguage( text, dm->mainWindow()->languageProfileMap() );

    // update the file record with the information about the language
    dm->updateLanguageOfFile( fileId, language );

    m_totalFullTexts++;
    m_catalog->setFullTexts( m_totalFullTexts );
    m_totalExtractions++;
    m_catalog->setExtractions( m_totalExtractions );
}

void KatInfoExtractor::noFullText( const KFileItem* kfileitem ) {
    m_totalExtractions++;
    m_catalog->setExtractions( m_totalExtractions );
}

void KatInfoExtractor::resultThumbnail( KIO::Job* job ) {
    m_jobs--;
    if ( m_jobs == 0 ) {
        startNextJob();
    }
}

void KatInfoExtractor::resultMetaInfo( KIO::Job* job ) {
    m_jobs--;
    if ( m_jobs == 0 ) {
        startNextJob();
    }
}

void KatInfoExtractor::resultFullText( KIO::Job* job ) {
    m_jobs--;
    if ( m_jobs == 0 ) {
        startNextJob();
    }
}

void KatInfoExtractor::slotExtractionAborted() {
    KIO::Job::slotResult( this );
}

void KatInfoExtractor::virtual_hook( int id, void* data ) {
    KIO::Job::virtual_hook( id, data );
}

#include "katinfoextractor.moc"
